@@ -9,8 +9,11 @@ use futures_util::stream::{StreamExt, once, iter};
 use futures_util::FutureExt;
 
 use fluvio_spu_schema::server::stream_fetch::{
-    DefaultStreamFetchRequest, DefaultStreamFetchResponse, SmartStreamPayload, SmartStreamWasm,
-    SmartStreamKind, WASM_MODULE_V2_API, GZIP_WASM_API,
+    DefaultStreamFetchRequest, DefaultStreamFetchResponse, GZIP_WASM_API, SMART_MODULE_API,
+    LegacySmartModulePayload, SmartModuleWasmCompressed, WASM_MODULE_API, WASM_MODULE_V2_API,
+};
+pub use fluvio_spu_schema::server::stream_fetch::{
+    SmartModuleInvocation, SmartModuleInvocationWasm, SmartModuleKind, DerivedStreamInvocation,
 };
 use dataplane::Isolation;
 use dataplane::ReplicaKey;
@@ -20,15 +23,16 @@ use dataplane::fetch::FetchPartition;
 use dataplane::fetch::FetchableTopic;
 use dataplane::fetch::FetchablePartitionResponse;
 use dataplane::record::RecordSet;
-use dataplane::record::Record as DefaultRecord;
 use dataplane::batch::Batch;
 use fluvio_types::event::offsets::OffsetPublisher;
 
 use crate::FluvioError;
 use crate::offset::{Offset, fetch_offsets};
-use crate::spu::SpuPool;
+use crate::spu::{SpuDirectory, SpuPool};
 use derive_builder::Builder;
 
+pub use dataplane::record::ConsumerRecord as Record;
+
 /// An interface for consuming events from a particular partition
 ///
 /// There are two ways to consume events: by "fetching" events
@@ -57,14 +61,17 @@ use derive_builder::Builder;
 /// [`Offset`]: struct.Offset.html
 /// [`partition_consumer`]: struct.Fluvio.html#method.partition_consumer
 /// [`Fluvio`]: struct.Fluvio.html
-pub struct PartitionConsumer {
+pub struct PartitionConsumer<P = SpuPool> {
     topic: String,
     partition: i32,
-    pool: Arc<SpuPool>,
+    pool: Arc<P>,
 }
 
-impl PartitionConsumer {
-    pub(crate) fn new(topic: String, partition: i32, pool: Arc<SpuPool>) -> Self {
+impl<P> PartitionConsumer<P>
+where
+    P: SpuDirectory,
+{
+    pub fn new(topic: String, partition: i32, pool: Arc<P>) -> Self {
         Self {
             topic,
             partition,
@@ -252,7 +259,7 @@ impl PartitionConsumer {
     pub async fn stream(
         &self,
         offset: Offset,
-    ) -> Result<impl Stream<Item = Result<Record, FluvioError>>, FluvioError> {
+    ) -> Result<impl Stream<Item = Result<Record, ErrorCode>>, FluvioError> {
         let config = ConsumerConfig::builder().build()?;
         let stream = self.stream_with_config(offset, config).await?;
 
@@ -302,7 +309,7 @@ impl PartitionConsumer {
         &self,
         offset: Offset,
         config: ConsumerConfig,
-    ) -> Result<impl Stream<Item = Result<Record, FluvioError>>, FluvioError> {
+    ) -> Result<impl Stream<Item = Result<Record, ErrorCode>>, FluvioError> {
         let stream = self.stream_batches_with_config(offset, config).await?;
         let partition = self.partition;
         let flattened =
@@ -356,7 +363,7 @@ impl PartitionConsumer {
         &self,
         offset: Offset,
         config: ConsumerConfig,
-    ) -> Result<impl Stream<Item = Result<Batch, FluvioError>>, FluvioError> {
+    ) -> Result<impl Stream<Item = Result<Batch, ErrorCode>>, FluvioError> {
         let stream = self.request_stream(offset, config).await?;
         let flattened = stream.flat_map(|batch_result: Result<DefaultStreamFetchResponse, _>| {
             let response = match batch_result {
@@ -373,12 +380,7 @@ impl PartitionConsumer {
                 let code = response.partition.error_code;
                 match code {
                     ErrorCode::None => None,
-                    ErrorCode::SmartStreamError(error) => {
-                        Some(Err(FluvioError::SmartStream(error)))
-                    }
-                    _ => Some(Err(FluvioError::AdminApi(
-                        fluvio_sc_schema::ApiError::Code(code, None),
-                    ))),
+                    _ => Some(Err(code)),
                 }
             };
 
@@ -397,11 +399,10 @@ impl PartitionConsumer {
         &self,
         offset: Offset,
         config: ConsumerConfig,
-    ) -> Result<impl Stream<Item = Result<DefaultStreamFetchResponse, FluvioError>>, FluvioError>
+    ) -> Result<impl Stream<Item = Result<DefaultStreamFetchResponse, ErrorCode>>, FluvioError>
     {
         use fluvio_future::task::spawn;
         use futures_util::stream::empty;
-        use fluvio_spu_schema::server::stream_fetch::WASM_MODULE_API;
         use fluvio_protocol::api::Request;
 
         let replica = ReplicaKey::new(&self.topic, self.partition);
@@ -426,31 +427,39 @@ impl PartitionConsumer {
         // add wasm module if SPU supports it
         let stream_fetch_version = serial_socket
             .versions()
-            .lookup_version(DefaultStreamFetchRequest::API_KEY)
+            .lookup_version(
+                DefaultStreamFetchRequest::API_KEY,
+                DefaultStreamFetchRequest::DEFAULT_API_VERSION,
+            )
             .unwrap_or((WASM_MODULE_API - 1) as i16);
+        debug!(%stream_fetch_version, "stream_fetch_version");
 
-        if let Some(mut module) = config.wasm_module {
-            if stream_fetch_version < WASM_MODULE_API as i16 {
-                return Err(FluvioError::Other("SPU does not support WASM".to_owned()));
-            }
+        stream_request.derivedstream = config.derivedstream;
 
-            if stream_fetch_version < WASM_MODULE_V2_API as i16 {
-                // SmartStream V1
-                debug!("Using WASM V1 API");
-                let wasm = module.wasm.get_raw()?;
-                stream_request.wasm_module = wasm.into_owned();
-            } else {
-                // SmartStream V2
-                debug!("Using WASM V2 API");
-                if stream_fetch_version < GZIP_WASM_API as i16 {
-                    module.wasm.to_raw()?;
+        if let Some(smartmodule) = config.smartmodule {
+            if stream_fetch_version < SMART_MODULE_API as i16 {
+                if let SmartModuleInvocationWasm::AdHoc(wasm) = smartmodule.wasm {
+                    let legacy_module = LegacySmartModulePayload {
+                        wasm: SmartModuleWasmCompressed::Gzip(wasm),
+                        kind: smartmodule.kind,
+                        params: smartmodule.params,
+                    };
+                    legacy_set_wasm(stream_fetch_version, &mut stream_request, legacy_module)?;
                 } else {
-                    debug!("Using compressed WASM API");
-                    module.wasm.to_gzip()?;
+                    return Err(FluvioError::Other(
+                        "SPU does not support persistent WASM".to_owned(),
+                    ));
                 }
-                stream_request.wasm_payload = Some(module);
+            } else {
+                debug!("Using persistent WASM API");
+                stream_request.smartmodule = Some(smartmodule);
             }
         }
+
+        if let Some(module) = config.wasm_module {
+            legacy_set_wasm(stream_fetch_version, &mut stream_request, module)?;
+        }
+
         let mut stream = self
             .pool
             .create_stream_with_version(&replica, stream_request, stream_fetch_version)
@@ -518,7 +527,10 @@ impl PartitionConsumer {
                         }
                         response
                     })
-                    .map_err(|e| e.into())
+                    .map_err(|e| {
+                        error!(?e, "error in stream");
+                        ErrorCode::Other(e.to_string())
+                    })
                 });
                 Either::Left(
                     iter(vec![Ok(response)])
@@ -539,6 +551,33 @@ impl PartitionConsumer {
     }
 }
 
+fn legacy_set_wasm(
+    stream_fetch_version: i16,
+    stream_request: &mut DefaultStreamFetchRequest,
+    mut module: LegacySmartModulePayload,
+) -> Result<(), FluvioError> {
+    if stream_fetch_version < WASM_MODULE_API as i16 {
+        return Err(FluvioError::Other("SPU does not support WASM".to_owned()));
+    }
+
+    if stream_fetch_version < WASM_MODULE_V2_API as i16 {
+        debug!("Using WASM V1 API");
+        let wasm = module.wasm.get_raw()?;
+        stream_request.wasm_module = wasm.into_owned();
+    } else {
+        debug!("Using WASM V2 API");
+        if stream_fetch_version < GZIP_WASM_API as i16 {
+            module.wasm.to_raw()?;
+        } else {
+            debug!("Using compressed WASM API");
+            module.wasm.to_gzip()?;
+        }
+        stream_request.wasm_payload = Some(module);
+    }
+
+    Ok(())
+}
+
 /// Wrap an inner record stream and only stream until a given number of records have been fetched.
 ///
 /// This is used for "disable continuous" mode. In this mode, we first make a FetchOffsetPartitionResponse
@@ -555,7 +594,7 @@ struct TakeRecords<S> {
 
 impl<S> TakeRecords<S>
 where
-    S: Stream<Item = Result<DefaultStreamFetchResponse, FluvioError>> + std::marker::Unpin,
+    S: Stream<Item = Result<DefaultStreamFetchResponse, ErrorCode>> + std::marker::Unpin,
 {
     pub fn new(stream: S, until: i64) -> Self {
         Self {
@@ -567,7 +606,7 @@ where
 
 impl<S> Stream for TakeRecords<S>
 where
-    S: Stream<Item = Result<DefaultStreamFetchResponse, FluvioError>> + std::marker::Unpin,
+    S: Stream<Item = Result<DefaultStreamFetchResponse, ErrorCode>> + std::marker::Unpin,
 {
     type Item = S::Item;
 
@@ -680,7 +719,11 @@ pub struct ConsumerConfig {
     #[builder(default)]
     pub(crate) isolation: Isolation,
     #[builder(private, default, setter(into, strip_option))]
-    pub(crate) wasm_module: Option<SmartStreamPayload>,
+    pub(crate) wasm_module: Option<LegacySmartModulePayload>,
+    #[builder(default)]
+    pub(crate) smartmodule: Option<SmartModuleInvocation>,
+    #[builder(default)]
+    pub(crate) derivedstream: Option<DerivedStreamInvocation>,
 }
 
 impl ConsumerConfig {
@@ -697,57 +740,76 @@ impl ConsumerConfigBuilder {
         Ok(config)
     }
 
-    /// Adds a SmartStream filter to this ConsumerConfig
+    /// Adds a SmartModule filter to this ConsumerConfig
+    #[deprecated(note = "Use 'smartmodule' instead", since = "0.9.11")]
     pub fn wasm_filter<T: Into<Vec<u8>>>(
         &mut self,
         filter: T,
         params: BTreeMap<String, String>,
     ) -> &mut Self {
-        self.wasm_module(SmartStreamPayload {
-            wasm: SmartStreamWasm::Raw(filter.into()),
-            kind: SmartStreamKind::Filter,
+        self.wasm_module(LegacySmartModulePayload {
+            wasm: SmartModuleWasmCompressed::Raw(filter.into()),
+            kind: SmartModuleKind::Filter,
             params: params.into(),
         });
         self
     }
 
-    /// Adds a SmartStream map to this ConsumerConfig
+    /// Adds a SmartModule map to this ConsumerConfig
+    #[deprecated(note = "Use 'smartmodule' instead", since = "0.9.11")]
     pub fn wasm_map<T: Into<Vec<u8>>>(
         &mut self,
         map: T,
         params: BTreeMap<String, String>,
     ) -> &mut Self {
-        self.wasm_module(SmartStreamPayload {
-            wasm: SmartStreamWasm::Raw(map.into()),
-            kind: SmartStreamKind::Map,
+        self.wasm_module(LegacySmartModulePayload {
+            wasm: SmartModuleWasmCompressed::Raw(map.into()),
+            kind: SmartModuleKind::Map,
             params: params.into(),
         });
         self
     }
 
-    /// Adds a SmartStream array_map to this ConsumerConfig
+    /// Adds a SmartModule filter_map to this ConsumerConfig
+    #[deprecated(note = "Use 'smartmodule' instead", since = "0.9.11")]
+    pub fn wasm_filter_map<T: Into<Vec<u8>>>(
+        &mut self,
+        filter_map: T,
+        params: BTreeMap<String, String>,
+    ) -> &mut Self {
+        self.wasm_module(LegacySmartModulePayload {
+            wasm: SmartModuleWasmCompressed::Raw(filter_map.into()),
+            kind: SmartModuleKind::FilterMap,
+            params: params.into(),
+        });
+        self
+    }
+
+    /// Adds a SmartModule array_map to this ConsumerConfig
+    #[deprecated(note = "Use 'smartmodule' instead", since = "0.9.11")]
     pub fn wasm_array_map<T: Into<Vec<u8>>>(
         &mut self,
         array_map: T,
         params: BTreeMap<String, String>,
     ) -> &mut Self {
-        self.wasm_module(SmartStreamPayload {
-            wasm: SmartStreamWasm::Raw(array_map.into()),
-            kind: SmartStreamKind::ArrayMap,
+        self.wasm_module(LegacySmartModulePayload {
+            wasm: SmartModuleWasmCompressed::Raw(array_map.into()),
+            kind: SmartModuleKind::ArrayMap,
             params: params.into(),
         })
     }
 
     /// Set a WASM aggregator function and initial accumulator value
+    #[deprecated(note = "Use 'smartmodule' instead", since = "0.9.11")]
     pub fn wasm_aggregate<T: Into<Vec<u8>>, U: Into<Vec<u8>>>(
         &mut self,
         aggregate: T,
         accumulator: U,
         params: BTreeMap<String, String>,
     ) -> &mut Self {
-        self.wasm_module(SmartStreamPayload {
-            wasm: SmartStreamWasm::Raw(aggregate.into()),
-            kind: SmartStreamKind::Aggregate {
+        self.wasm_module(LegacySmartModulePayload {
+            wasm: SmartModuleWasmCompressed::Raw(aggregate.into()),
+            kind: SmartModuleKind::Aggregate {
                 accumulator: accumulator.into(),
             },
             params: params.into(),
@@ -832,7 +894,7 @@ impl MultiplePartitionConsumer {
     pub async fn stream(
         &self,
         offset: Offset,
-    ) -> Result<impl Stream<Item = Result<Record, FluvioError>>, FluvioError> {
+    ) -> Result<impl Stream<Item = Result<Record, ErrorCode>>, FluvioError> {
         let config = ConsumerConfig::builder().build()?;
         let stream = self.stream_with_config(offset, config).await?;
 
@@ -882,7 +944,7 @@ impl MultiplePartitionConsumer {
         &self,
         offset: Offset,
         config: ConsumerConfig,
-    ) -> Result<impl Stream<Item = Result<Record, FluvioError>>, FluvioError> {
+    ) -> Result<impl Stream<Item = Result<Record, ErrorCode>>, FluvioError> {
         let consumers = self
             .strategy
             .selection(self.pool.clone())
@@ -903,49 +965,6 @@ impl MultiplePartitionConsumer {
     }
 }
 
-/// The individual record for a given stream.
-pub struct Record {
-    /// The offset of this Record into its partition
-    offset: i64,
-    /// The partition where this Record is stored
-    partition: i32,
-    /// The Record contents
-    record: DefaultRecord,
-}
-
-impl Record {
-    /// The offset from the initial offset for a given stream.
-    pub fn offset(&self) -> i64 {
-        self.offset
-    }
-
-    /// The partition where this Record is stored.
-    pub fn partition(&self) -> i32 {
-        self.partition
-    }
-
-    /// Returns the contents of this Record's key, if it exists
-    pub fn key(&self) -> Option<&[u8]> {
-        self.record.key().map(|it| it.as_ref())
-    }
-
-    /// Returns the contents of this Record's value
-    pub fn value(&self) -> &[u8] {
-        self.record.value().as_ref()
-    }
-
-    /// Returns the inner representation of the Record
-    pub fn into_inner(self) -> DefaultRecord {
-        self.record
-    }
-}
-
-impl AsRef<[u8]> for Record {
-    fn as_ref(&self) -> &[u8] {
-        self.value()
-    }
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
